#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <limits.h>

#define MAX_SIZE 1024
int n = 0, m = 0, total_lambdas = 0;

#define MAX(a, b) ((a) > (b) ? (a) : (b))
#define MIN(a, b) ((a) < (b) ? (a) : (b))

typedef struct Pos {
    int x, y;
} Pos;

#define DEAD		0x01
#define BLOCKED		0x02
#define STUCK		0x04
#define CHECKED		0x08
#define FALLING		0x10

#define DEBUG(args...)
// #define DEBUG(args...) fprintf(stderr, args);

typedef struct State {
    char *map;
    char *road;
    int lambdas;
    int distance;
    int waterproof;
    int razors;
    int water;
    int moves;
    int score;
    int rx, ry;
    int flags;
} State;

State best;

int pos_index(int x, int y) {
    return (x - 1) * m + (y - 1);
}

void set(State *state, int x, int y, char val) {
    state->map[pos_index(x, y)] = val;
}

char get(State *state, int x, int y) {
    return state->map[pos_index(x, y)];
}

char safe_get(State *state, int x, int y) {
    return (x > n || x < 1 || y > m || y < 1) ? '#' : get(state, x, y);
}

void safe_set(State *state, int x, int y, char val) {
    if (x <= n && x >= 1 && y <= m && y >= 1) {
	set(state, x, y, val);
    }
}

char *empty(void) {
    char *str = malloc(1);
    str[0] = 0;
    return str;
}

State dummy(void) {
    return (State) { score: 0,	    
		     flags: 0,
	             moves: 0,
	             water: 0,
	             razors: 0,
	             lambdas: 0,
	             waterproof: 10,
	             distance: 2 * MAX_SIZE,
	             road: empty() };
}

State init_state(State *old) {
    State ret = old ? *old : dummy();
    ret.map = malloc(n * m);
    return ret;
}

int is_valid(State *state) {
    return (state->road != NULL);
}

void copy_state(State *dst, State *src) {
    if (!is_valid(dst)) {
	*dst = init_state(NULL);
    }
    char *map = dst->map;
    free(dst->road);
    *dst = *src;
    dst->road = malloc(src->moves + 1);
    strcpy(dst->road, src->road);
    memcpy(map, src->map, n * m);
    dst->map = map;
}

void free_state(State *old) {
    if (old->road) {
	free(old->map);
	free(old->road);
	old->road = NULL;
    }
}

int is_rock(char c) {
    return (c == '*' || c == '@');
}

int blocking(State *state, int x, int y) {
    char c = safe_get(state, x, y);
    return is_rock(c) || c == '#';
}

int check_if_lift_blocked_from_top(State *state, int lx, int ly) {
    if (ly < m) {
	char c1 = safe_get(state, lx, ly + 1);
	char c2 = safe_get(state, lx, ly + 2);
	// just common rock, lambda rock will shatter by falling
	if ((c2 == '*') && (c1 == ' '  || c1 == '\\' || c1 == 'R')) {
	    return 1;
	}
	else {
	    return check_if_lift_blocked_from_top(state, lx, ly + 1);
	}
    }
    return 0;
}

void check_if_lift_blocked(State *state, int lx, int ly) {
    if (get(state, lx, ly) == 'L'
	&& blocking(state, lx + 1, ly)
	&& blocking(state, lx - 1, ly)
	&& blocking(state, lx, ly - 1)
	&& (blocking(state, lx, ly + 1)
	    || check_if_lift_blocked_from_top(state, lx, ly))) {
	state->flags |= BLOCKED;
    }
}

int stuck(State *state, int x, int y, int dir) {
    return (safe_get(state, x + dir * 1, y) == '#'
	    || (is_rock(safe_get(state, x + dir * 1, y))
		&& safe_get(state, x + dir * 2, y) != ' '));
} 

void check_if_robot_blocked(State *state) {
    int x = state->rx, y = state->ry;
    if (is_rock(safe_get(state, x, y + 1))
	&& stuck(state, x, y, 1)
	&& stuck(state, x, y, -1)) {
	state->flags |= STUCK;
    }
}

int growth = 25;
int flooding = 0;
char trampolines[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };

int trampoline_index(char c) {
    return (int) (c - 'A');
}

void update_distance(State *state, int x, int y) {
    char c = get(state, x, y);
    if (c == '\\' || (state->lambdas == 0 && c == 'L')) {
	int manhattan = abs(state->rx - x) + abs(state->ry - y);
	state->distance = MIN(state->distance, manhattan);
    }
}

State read_map(void) {
    int i, j;
    char *str;
    State ret;
    char name[256];
    char map[MAX_SIZE][MAX_SIZE];
    
    do {
	str = fgets(map[m], MAX_SIZE, stdin);
	if (str) {
	    int len = strlen(str);
	    if (len >= 2 && str[len - 2] == 0xd) {
		len = len - 1;
	    }
	    if (strcmp(str, "\n") == 0) {
		break;
	    }
	    n = MAX(n, len);
	    m++;
	}
    } while (str);

    int count;
    ret = init_state(NULL);
    do {
	count = scanf("%s", name);
	if (count > 0) {
	    if (strcmp(name, "Water") == 0) {
		count = scanf("%d\n", &ret.water);
	    }
	    if (strcmp(name, "Flooding") == 0) {
		count = scanf("%d\n", &flooding);
	    }
	    if (strcmp(name, "Waterproof") == 0) {
		count = scanf("%d\n", &ret.waterproof);
	    }
	    if (strcmp(name, "Razors") == 0) {
		count = scanf("%d\n", &ret.razors);
	    }
	    if (strcmp(name, "Growth") == 0) {
		count = scanf("%d\n", &growth);
	    }
	    if (strcmp(name, "Trampoline") == 0) {
		char letter = 0, number = 0;
		count = scanf(" %c %s %c\n", &letter, name, &number);
		trampolines[trampoline_index(letter)] = number;
	    }
	}
    } while (count > 0);	    
    
    for (i = 0; i < m; i++) {
	for (j = strlen(map[i]) - 1; j < n; j++) {
	    map[i][j] = ' ';
	}
	map[i][n] = 0;
    } 

    int x, y;
    int lx = 1, ly = 1;
    for (x = 1; x <= n; x++) {
	for (y = 1; y <= m; y++) {
 	    char c = map[m - y][x - 1];
	    set(&ret, x, y, c);
	    if (c == '\\' || c == '@') {
		total_lambdas++;
	    }
	    if (c == 'R') {
		ret.rx = x;
		ret.ry = y;
	    }
	    if (c == 'L') {
		lx = x;
		ly = y;
	    }
	    update_distance(&ret, x, y);
	}
    }

    check_if_lift_blocked(&ret, lx, ly);
    ret.lambdas = total_lambdas;

    return ret;
}

void set_rock(State *state, int x, int y, char type) {
    char c = safe_get(state, x, y - 1);
    set(state, x, y, (c != ' ' && type == '@') ? '\\' : type);
    if (type == '*') {
	state->flags |= FALLING;
    }
    if (c == 'R') {
	state->flags |= DEAD;
    }
}

void advance_beard(State *new, State *old, int x, int y) {
    if (safe_get(new, x - 1, y    ) == ' ') safe_set(new, x - 1, y    , 'W');
    if (safe_get(new, x - 1, y - 1) == ' ') safe_set(new, x - 1, y - 1, 'W');
    if (safe_get(new, x    , y - 1) == ' ') safe_set(new, x    , y - 1, 'W');
    if (safe_get(new, x + 1, y - 1) == ' ') safe_set(new, x + 1, y - 1, 'W');

    if (safe_get(old, x + 1, y    ) == ' ') safe_set(old, x + 1, y    , 'w');
    if (safe_get(old, x - 1, y + 1) == ' ') safe_set(old, x - 1, y + 1, 'w');
    if (safe_get(old, x    , y + 1) == ' ') safe_set(old, x    , y + 1, 'w');
    if (safe_get(old, x + 1, y + 1) == ' ') safe_set(old, x + 1, y + 1, 'w');
}

State update_map(State *old) {
    int x, y;
    int lx = 1, ly = 1;
    State new = init_state(old);
    new.distance = 2 * MAX_SIZE;
    new.flags = 0;
    int beard_advance = growth ? (new.moves % growth == 0) : 0;
    for (y = 1; y <= m; y++) {
	for (x = 1; x <= n; x++) {
	    char c = get(old, x, y);
	    if (c == 'W' && beard_advance) {
		advance_beard(&new, old, x, y);
	    }
	    if (c == 'w') {
		set(&new, x, y, 'W');
		continue;
	    }
	    // rock falls
	    if (y > 1 && is_rock(c) && get(old, x, y - 1) == ' ') {
		set_rock(&new, x, y - 1, c);
		set(&new, x, y, ' ');
		continue;
	    }
	    // rock slides to right
	    if (y > 1 && x < n && is_rock(c) 
			       && get(old, x + 1, y) == ' '
		               && get(old, x + 1, y - 1) == ' ') {
		char under = get(old, x, y - 1);
		if (is_rock(under) || under == '\\') {
		    set_rock(&new, x + 1, y - 1, c);
		    set(&new, x, y, ' ');
		    continue;
		}
	    }
	    // rock slides to left
	    if (y > 1 && x > 1 && is_rock(c)
			       && is_rock(get(old, x, y - 1))
		               && get(old, x - 1, y) == ' '
		               && get(old, x - 1, y - 1) == ' ') {
		set_rock(&new, x - 1, y - 1, c);
		set(&new, x, y, ' ');
		continue;
	    }
	    set(&new, x, y, c);
	    update_distance(&new, x, y);	    
	    if (c == 'L') {
		lx = x;
		ly = y;
	    }
	}
    }
    int water = new.water;
    if (flooding) water += new.moves / flooding;
    if (new.ry <= water) new.waterproof--;
    if (new.waterproof <= 0) new.flags |= DEAD;
    check_if_lift_blocked(&new, lx, ly);
    check_if_robot_blocked(&new);
    return new;
}

char move(int x, int y, int s) {
    if (x == 1) {
	return 'R';
    }
    else if (x == -1) {
	return 'L';
    }
    else if (y == 1) {
	return 'U';
    }
    else if (y == -1) {
	return 'D';
    }
    else if (s == 1) {
	return 'S';
    }
    else {
	return 'W';
    }
}

void print_score(State *state, int compensate) {
    int score = state->score;
    score += compensate * (total_lambdas - state->lambdas);
    DEBUG("score = %d\n", score);
}

void dump_state(State *state) {
    int x, y;
    for (y = m; y >= 1; y--) {
	for (x = 1; x <= n; x++) {
	    DEBUG("%c", get(state, x, y));
	}
	DEBUG("\n");
    }
    DEBUG("%s\n", state->road);
    DEBUG("\n");
}

void find_trapoline_target(State *state, char type, int *tx, int *ty) {    
    int i;
    int x, y;
    char target = trampolines[trampoline_index(type)];
    for (y = 1; y <= m; y++) {
	for (x = 1; x <= n; x++) {
	    if (get(state, x, y) == target) {
		*tx = x;
		*ty = y;
	    }
	    for (i = 0; i < 9; i++) {
		if (trampolines[i] == target && get(state, x, y) == 'A' + i) {
		    set(state, x, y, ' ');		    
		}
	    }
	}
    }
}

void shave_cell(State *state, int dx, int dy) {
    if (safe_get(state, state->rx + dx, state->ry + dy) == 'W') {
	set(state, state->rx + dx, state->ry + dy, ' ');
    }
}

void shave_waddler(State *state) {
    shave_cell(state,  1,  1);
    shave_cell(state,  0,  1);
    shave_cell(state, -1,  1);
    shave_cell(state,  1,  0);
    shave_cell(state, -1,  0);
    shave_cell(state,  1, -1);
    shave_cell(state,  0, -1);
    shave_cell(state, -1, -1);
}

State advance(State *old, int x, int y, int shave) {
    char tmp_map[n * m];
    State tmp_state = *old;
    tmp_state.map = tmp_map;
    memcpy(tmp_map, old->map, n * m);
    int new_x = x + old->rx;
    int new_y = y + old->ry;
    int rock_x = new_x + x;
    tmp_state.score--;
    tmp_state.moves++;
    tmp_state.road = malloc(tmp_state.moves + 1);
    sprintf(tmp_state.road, "%s%c", old->road, move(x, y, shave));
    if (shave) {
	shave_waddler(&tmp_state);	
    }
    else if (new_x >= 1 && new_x <= n && new_y >= 1 && new_y <= m) {
	char target = get(old, new_x, new_y);
	int push_rock = is_rock(target) && x != 0;
	push_rock = push_rock && new_x >= 2 && new_x <= n - 1;
	push_rock = push_rock && get(old, rock_x, new_y) == ' ';
	int jump = (target >= 'A' && target <= 'I');
	if (target == '.' || target == ' ' || target == '\\' 
		|| push_rock || jump || target == '!') {
	    if (push_rock) {
		set(&tmp_state, rock_x, new_y, target);
	    }
	    set(&tmp_state, old->rx, old->ry, ' ');
	    if (jump) {
		set(&tmp_state, new_x, new_y, ' ');
		find_trapoline_target(&tmp_state, target, &new_x, &new_y);
	    }
	    set(&tmp_state, new_x, new_y, 'R');
	    tmp_state.rx = new_x;
	    tmp_state.ry = new_y;
	    if (target == '\\') {
		tmp_state.lambdas--;
		tmp_state.score += 25;		
	    }
	    if (target == '!') {
		tmp_state.razors++;
	    }
	}
	else if (target == 'L' && tmp_state.lambdas == 0) {
	    printf("%s\n", tmp_state.road);
	    print_score(&tmp_state, 50);
	    dump_state(&tmp_state);
	    exit(0);
	}
    }
    return update_map(&tmp_state);
}

#define MAX_ACTIVE_STATES 1024
State active[MAX_ACTIVE_STATES];
int current = 1;
int tolerance = 0;

int same_place(State *a, State *b) {
    return (a->rx == b->rx && a->ry == b->ry);
}

int has_better(State *state) {
    int i;
    for (i = 0; i < MAX_ACTIVE_STATES; i++) {
	if (is_valid(&active[i])
	    && same_place(&active[i], state)
	    && active[i].score - tolerance >= state->score) {
	    return 1;
	}
    }
    return 0;
}

int miss = 0;
void look(State *state, int x, int y, int s) {
    State new = advance(state, x, y, s);
    if ((new.flags & DEAD) || has_better(&new) || same_place(state, &new)) {
	free_state(&new);
	return;
    }
    if (is_valid(&active[current])) {
	free_state(&active[current]);
    }
    active[current] = new;
    miss = 0;
    if (new.score > best.score) {
	copy_state(&best, &new);
	tolerance = 0;
    }
    current = (current + 1) % MAX_ACTIVE_STATES;
}

char safe_get_near(State *state, int x, int y) {
    return safe_get(state, state->rx + x, state->ry + y);
}

int goodness(State *state) {
    int penalty = 0;
    if (!is_valid(state)) {
	return INT_MIN;
    }
    if (state->flags & STUCK) {
	penalty += 100;
    }
    if (state->flags & FALLING) {
	penalty += 5;
    }
    if (state->flags & BLOCKED) {
	penalty += 1000;
    }
    if (safe_get_near(state, 0, 1) == '*') {
	penalty += 10;
    }
    penalty -= (n + m) * (total_lambdas - state->lambdas);
    return state->razors * 10 + state->score - penalty - 2 * state->distance;
}

int find_best_new_state(int i) {
    i = (i + 1) % MAX_ACTIVE_STATES;
    int local_best = i;
    while (i != current) {
	int fresh = !(active[i].flags & CHECKED);
	if (fresh && goodness(&active[i]) > goodness(&active[local_best])) {
	    local_best = i;	    
	}
	i = (i + 1) % MAX_ACTIVE_STATES;
    }
    return local_best;
}

int check_waddler(State *state, int dx, int dy) {
    return (safe_get(state, state->rx + dx, state->ry + dy) == 'W');
}

int beard_move_reasonable(State *state) {
    return (state->razors > 0 
	    && (check_waddler(state, 1, 0)
		|| check_waddler(state, -1,  0)
		|| check_waddler(state,  1,  1)
		|| check_waddler(state,  0,  1)
		|| check_waddler(state, -1,  1)
		|| check_waddler(state,  1, -1)
		|| check_waddler(state,  0, -1)
		|| check_waddler(state, -1, -1)));
}

void explore(int i) {
    State *state = &active[i];
    miss = miss + 1;
    if (is_valid(state) && !(state->flags & CHECKED)) {
	state->flags |= CHECKED;
	look(state,  1,  0,  0);
	look(state,  0,  1,  0);
	look(state, -1,  0,  0);
	look(state,  0, -1,  0);
	if (beard_move_reasonable(state)) {
	    look(state,  0,  0,  1);
	}
    }
    if (miss > MAX_ACTIVE_STATES) {
	tolerance = tolerance + 1;
	for (i = 0; i < MAX_ACTIVE_STATES; i++) {
	    active[i].flags &= ~CHECKED;
	    current = i;
	}
	miss = 0;
    }
    explore(find_best_new_state(i));
}

void handle_SIGINT(int sig) {
    printf("%s\n", best.road);
    print_score(&best, 25);
    dump_state(&best);
    exit(0);
}

int main(int argc, char **argv) {
    int i;
    for (i = 0; i < MAX_ACTIVE_STATES; i++) {
	active[i].road = NULL;
    }    
    active[0] = read_map();
    best = init_state(NULL);
    copy_state(&best, &active[0]);
    signal(SIGINT, handle_SIGINT);
    explore(0);
    return 0;
}
